@@ -34,6 +34,8 @@ typedef struct {
 } SinkFunc;
 static GArray *sink_funcs = NULL;
 
+PyTypeObject *PyGObject_MetaType = NULL;
+
 /**
  * pygobject_sink:
  * @obj: a GObject
@@ -514,7 +516,7 @@ pygobject_register_class(PyObject *dict, const gchar *type_name,
     } else
         bases = runtime_bases;
 
-    type->ob_type = &PyGObject_MetaType;
+    type->ob_type = PyGObject_MetaType;
     type->tp_bases = bases;
     if (G_LIKELY(bases)) {
         type->tp_base = (PyTypeObject *)PyTuple_GetItem(bases, 0);
@@ -737,9 +739,10 @@ pygobject_lookup_class(GType gtype)
 }
 
 /**
- * pygobject_new:
+ * pygobject_new_full:
  * @obj: a GObject instance.
  * @sink: whether to sink any floating reference found on the GObject.
+ * @g_class: the GObjectClass
  *
  * This function gets a reference to a wrapper for the given GObject
  * instance.  If a wrapper has already been created, a new reference
@@ -749,7 +752,7 @@ pygobject_lookup_class(GType gtype)
  * Returns: a reference to the wrapper for the GObject.
  */
 PyObject *
-pygobject_new_full(GObject *obj, gboolean sink)
+pygobject_new_full(GObject *obj, gboolean sink, gpointer g_class)
 {
     PyGObject *self;
 
@@ -764,7 +767,11 @@ pygobject_new_full(GObject *obj, gboolean sink)
 	Py_INCREF(self);
     } else {
 	/* create wrapper */
-	PyTypeObject *tp = pygobject_lookup_class(G_OBJECT_TYPE(obj));
+	PyTypeObject *tp;
+        if (g_class)
+            tp = pygobject_lookup_class(G_OBJECT_CLASS_TYPE(g_class));
+        else
+            tp = pygobject_lookup_class(G_OBJECT_TYPE(obj));
         /* need to bump type refcount if created with
            pygobject_new_with_interfaces(). fixes bug #141042 */
         if (tp->tp_flags & Py_TPFLAGS_HEAPTYPE)
@@ -790,10 +797,11 @@ pygobject_new_full(GObject *obj, gboolean sink)
     return (PyObject *)self;
 }
 
+
 PyObject *
 pygobject_new(GObject *obj)
 {
-    return pygobject_new_full(obj, TRUE);
+    return pygobject_new_full(obj, TRUE, NULL);
 }
 
 static void
@@ -1008,7 +1016,7 @@ pygobject_init(PyGObject *self, PyObject *args, PyObject *kwargs)
 	    pspec = g_object_class_find_property (class, key_str);
 	    if (!pspec) {
 		PyErr_Format(PyExc_TypeError,
-			     "gobject `%s' doesn't support property `%s'",
+			     "object of type `%s' doesn't support property `%s'",
 			     g_type_name(object_type), key_str);
 		goto cleanup;
 	    }
@@ -1071,8 +1079,9 @@ pygobject_get_property(PyGObject *self, PyObject *args)
     pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(self->obj),
 					 param_name);
     if (!pspec) {
-	PyErr_SetString(PyExc_TypeError,
-			"the object does not support the given parameter");
+	PyErr_Format(PyExc_TypeError,
+		     "object of type `%s' does not have property `%s'",
+		     g_type_name(G_OBJECT_TYPE(self->obj)), param_name);
 	return NULL;
     }
     if (!(pspec->flags & G_PARAM_READABLE)) {
@@ -1103,8 +1112,9 @@ pygobject_set_property(PyGObject *self, PyObject *args)
     pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(self->obj),
 					 param_name);
     if (!pspec) {
-	PyErr_SetString(PyExc_TypeError,
-			"the object does not support the given parameter");
+	PyErr_Format(PyExc_TypeError,
+		     "object of type `%s' does not have property `%s'",
+		     g_type_name(G_OBJECT_TYPE(self->obj)), param_name);
 	return NULL;
     }
     
@@ -1628,6 +1638,102 @@ pygobject_chain_from_overridden(PyGObject *self, PyObject *args)
     return py_ret;
 }
 
+static PyObject *
+pygobject_disconnect_by_func(PyGObject *self, PyObject *args)
+{
+    PyObject *pyfunc = NULL;
+    GClosure *closure = NULL;
+    guint retval;
+    
+    CHECK_GOBJECT(self);
+
+    if (!PyArg_ParseTuple(args, "O:GObject.disconnect_by_func", &pyfunc))
+	return NULL;
+
+    if (!PyCallable_Check(pyfunc)) {
+	PyErr_SetString(PyExc_TypeError, "first argument must be callable");
+	return NULL;
+    }
+
+    closure = gclosure_from_pyfunc(self, pyfunc);
+    if (!closure) {
+	PyErr_Format(PyExc_TypeError, "nothing connected to %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)pyfunc)));
+	return NULL;
+    }
+    
+    retval = g_signal_handlers_disconnect_matched(self->obj,
+						  G_SIGNAL_MATCH_CLOSURE,
+						  0, 0,
+						  closure,
+						  NULL, NULL);
+    return PyInt_FromLong(retval);
+}
+
+static PyObject *
+pygobject_handler_block_by_func(PyGObject *self, PyObject *args)
+{
+    PyObject *pyfunc = NULL;
+    GClosure *closure = NULL;
+    guint retval;
+    
+    CHECK_GOBJECT(self);
+
+    if (!PyArg_ParseTuple(args, "O:GObject.handler_block_by_func", &pyfunc))
+	return NULL;
+
+    if (!PyCallable_Check(pyfunc)) {
+	PyErr_SetString(PyExc_TypeError, "first argument must be callable");
+	return NULL;
+    }
+
+    closure = gclosure_from_pyfunc(self, pyfunc);
+    if (!closure) {
+	PyErr_Format(PyExc_TypeError, "nothing connected to %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)pyfunc)));
+	return NULL;
+    }
+    
+    retval = g_signal_handlers_block_matched(self->obj,
+					     G_SIGNAL_MATCH_CLOSURE,
+					     0, 0,
+					     closure,
+					     NULL, NULL);
+    return PyInt_FromLong(retval);
+}
+
+static PyObject *
+pygobject_handler_unblock_by_func(PyGObject *self, PyObject *args)
+{
+    PyObject *pyfunc = NULL;
+    GClosure *closure = NULL;
+    guint retval;
+    
+    CHECK_GOBJECT(self);
+
+    if (!PyArg_ParseTuple(args, "O:GObject.handler_unblock_by_func", &pyfunc))
+	return NULL;
+
+    if (!PyCallable_Check(pyfunc)) {
+	PyErr_SetString(PyExc_TypeError, "first argument must be callable");
+	return NULL;
+    }
+
+    closure = gclosure_from_pyfunc(self, pyfunc);
+    if (!closure) {
+	PyErr_Format(PyExc_TypeError, "nothing connected to %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)pyfunc)));
+	return NULL;
+    }
+    
+    retval = g_signal_handlers_unblock_matched(self->obj,
+					       G_SIGNAL_MATCH_CLOSURE,
+					       0, 0,
+					       closure,
+					       NULL, NULL);
+    return PyInt_FromLong(retval);
+}
+
 static PyMethodDef pygobject_methods[] = {
     { "__gobject_init__", (PyCFunction)pygobject__gobject_init__,
       METH_VARARGS|METH_KEYWORDS },
@@ -1643,10 +1749,13 @@ static PyMethodDef pygobject_methods[] = {
     { "connect_object", (PyCFunction)pygobject_connect_object, METH_VARARGS },
     { "connect_object_after", (PyCFunction)pygobject_connect_object_after, METH_VARARGS },
     { "disconnect", (PyCFunction)pygobject_disconnect, METH_VARARGS },
+    { "disconnect_by_func", (PyCFunction)pygobject_disconnect_by_func, METH_VARARGS },
     { "handler_disconnect", (PyCFunction)pygobject_disconnect, METH_VARARGS },
     { "handler_is_connected", (PyCFunction)pygobject_handler_is_connected, METH_VARARGS },
     { "handler_block", (PyCFunction)pygobject_handler_block, METH_VARARGS },
     { "handler_unblock", (PyCFunction)pygobject_handler_unblock,METH_VARARGS },
+    { "handler_block_by_func", (PyCFunction)pygobject_handler_block_by_func, METH_VARARGS },
+    { "handler_unblock_by_func", (PyCFunction)pygobject_handler_unblock_by_func, METH_VARARGS },
     { "emit", (PyCFunction)pygobject_emit, METH_VARARGS },
     { "stop_emission", (PyCFunction)pygobject_stop_emission, METH_VARARGS },
     { "emit_stop_by_name", (PyCFunction)pygobject_stop_emission,METH_VARARGS },
@@ -1726,91 +1835,3 @@ PyTypeObject PyGObject_Type = {
 };
 
 
-static int
-pygobjectmeta_register(PyTypeObject *subtype)
-{
-    PyObject *pytype_name;
-    const char *type_name = NULL;
-    int retval = 0;
-    
-    /* Maybe the type doesn't really need to registered? */
-    if (subtype->tp_dict == NULL)
-	goto out;
-	
-    if (!(PyDict_GetItemString(subtype->tp_dict, "__gtype_name__")
-          || PyDict_GetItemString(subtype->tp_dict, "__gproperties__")
-          || PyDict_GetItemString(subtype->tp_dict, "__gsignals__")))
-	goto out;
-
-    /* If it's already registered, skip registration */
-    if (PyDict_GetItemString(subtype->tp_dict, "__gtype__"))
-	goto out;
-
-    pytype_name = PyDict_GetItemString(subtype->tp_dict, "__gtype_name__");
-    if (pytype_name)
-	type_name = PyString_AsString(pytype_name);
-    retval = pyg_type_register(subtype, type_name);
-
-out:
-    return retval;
-}
-
-/* -------- GObject MetaClass -----------*/
-static int
-pygobjectmeta_init(PyTypeObject *subtype, PyObject *args, PyObject *kwargs)
-{
-    if (PyType_Type.tp_init((PyObject *) subtype, args, kwargs))
-        return -1;
-    
-    return pygobjectmeta_register(subtype);
-}
-
-
-PyTypeObject PyGObject_MetaType = {
-    PyObject_HEAD_INIT(NULL)
-    0,					/* ob_size */
-    "gobject.GObjectMeta",		/* tp_name */
-    0,					/* tp_basicsize */
-    0,					/* tp_itemsize */
-    /* methods */
-    (destructor)0,			/* tp_dealloc */
-    (printfunc)0,			/* tp_print */
-    (getattrfunc)0,			/* tp_getattr */
-    (setattrfunc)0,			/* tp_setattr */
-    (cmpfunc)0,				/* tp_compare */
-    (reprfunc)0,			/* tp_repr */
-    0,					/* tp_as_number */
-    0,					/* tp_as_sequence */
-    0,					/* tp_as_mapping */
-    (hashfunc)0,			/* tp_hash */
-    (ternaryfunc)0,			/* tp_call */
-    (reprfunc)0,			/* tp_str */
-    (getattrofunc)0,			/* tp_getattro */
-    (setattrofunc)0,			/* tp_setattro */
-    0,					/* tp_as_buffer */
-    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_BASETYPE, /* tp_flags */
-    "GObject Metaclass -- automatically registers GObject subclasses "
-    "as new GType's", /* Documentation string */
-    (traverseproc)0,			/* tp_traverse */
-    (inquiry)0,				/* tp_clear */
-    (richcmpfunc)0,			/* tp_richcompare */
-    0,					/* tp_weaklistoffset */
-    (getiterfunc)0,			/* tp_iter */
-    (iternextfunc)0,			/* tp_iternext */
-    0,                                  /* tp_methods */
-    0,					/* tp_members */
-    0,					/* tp_getset */
-    NULL,				/* tp_base */
-    (PyObject *)0,			/* tp_dict */
-    0,					/* tp_descr_get */
-    0,					/* tp_descr_set */
-    0,					/* tp_dictoffset */
-    (initproc)pygobjectmeta_init,	/* tp_init */
-    (allocfunc)0,			/* tp_alloc */
-    (newfunc)0,				/* tp_new */
-    0,					/* tp_free */
-    (inquiry)0,				/* tp_is_gc */
-    NULL,				/* tp_bases */
-};
-
-
