@@ -577,7 +577,7 @@ pyg_type_lookup(GType type)
 
 void
 pyg_register_gtype_custom(GType gtype,
-		                  fromvaluefunc from_func,
+			  fromvaluefunc from_func,
                           tovaluefunc to_func)
 {
     PyGTypeMarshal *tm;
@@ -852,6 +852,17 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	else if (PySequence_Check(obj) &&
 		   G_VALUE_HOLDS(value, G_TYPE_VALUE_ARRAY))
 	    return pyg_value_array_from_pyobject(value, obj, NULL);
+	else if (PyString_Check(obj) &&
+                 G_VALUE_HOLDS(value, G_TYPE_GSTRING)) {
+            GString *string;
+            char *buffer;
+            int len;
+            if (PyString_AsStringAndSize(obj, &buffer, &len))
+                return -1;
+            string = g_string_new_len(buffer, len);
+	    g_value_set_boxed(value, string);
+            break;
+        }
 	else if ((bm = pyg_type_lookup(G_VALUE_TYPE(value))) != NULL)
 	    return bm->tovalue(value, obj);
 	else if (PyCObject_Check(obj))
@@ -1000,6 +1011,10 @@ pyg_value_as_pyobject(const GValue *value, gboolean copy_boxed)
 		PyList_SET_ITEM(ret, i, pyg_value_as_pyobject
                                 (array->values + i, copy_boxed));
 	    return ret;
+	} else if (G_VALUE_HOLDS(value, G_TYPE_GSTRING)) {
+	    GString *string = (GString *) g_value_get_boxed(value);
+	    PyObject *ret = PyString_FromStringAndSize(string->str, string->len);
+	    return ret;
 	}	    
 	bm = pyg_type_lookup(G_VALUE_TYPE(value));
 	if (bm) {
@@ -1306,6 +1321,21 @@ pyg_signal_class_closure_get(void)
     return closure;
 }
 
+GClosure *
+gclosure_from_pyfunc(PyGObject *object, PyObject *func)
+{
+    GSList *l;
+
+    for (l = object->closures; l; l = l->next) {
+	PyGClosure *pyclosure = l->data;
+	if (PyFunction_GetClosure(pyclosure->callback) ==
+		PyFunction_GetClosure(func)) {
+	    return (GClosure*)pyclosure;
+	}
+    }
+    return NULL;
+}
+
 /* ----- __doc__ descriptor for GObject and GInterface ----- */
 
 static void
